import Rank._
import Suit._
import HandVal._
class Hand(cardsAsStr: String) extends Object with Ordered[Hand] {
    protected val cards : List[(Rank.Value, Suit.Value)] = {
      val tmpStr = cardsAsStr.split(" ") 
      List.tabulate(5)(x => ( Rank.withName(tmpStr(x).charAt(0).toString), Suit.withName(tmpStr(x).charAt(1).toString) ) )
      /** the above line is equivalent to:
      List(   
           ( Rank.withName(tmpStr(0).charAt(0).toString), Suit.withName(tmpStr(0).charAt(1).toString) ),
           ( Rank.withName(tmpStr(1).charAt(0).toString), Suit.withName(tmpStr(1).charAt(1).toString) ),
           ( Rank.withName(tmpStr(2).charAt(0).toString), Suit.withName(tmpStr(2).charAt(1).toString) ),
           ( Rank.withName(tmpStr(3).charAt(0).toString), Suit.withName(tmpStr(3).charAt(1).toString) ),
           ( Rank.withName(tmpStr(4).charAt(0).toString), Suit.withName(tmpStr(4).charAt(1).toString) )
      )  */   
    }
    protected val handCalcList : List[(Int, Rank.Value)] = {   // aggregated representation of cards as basis for assessing the value of the hand
    												           // for example, a poker with 10s and a queen as single card implies handCalcList = ((4, T), (1, Q))
      val rankL= cards map(x => x._1)	// list containing only ranks
      val l= rankL map(x=> (rankL.count(y=> y == x), x))  // list with tuples (no. of occurences of particular rank, rank) 
      l.toSet.toList.sortWith(_ > _)	// remove duplicates (toSet) and sort resulting tuple list descending
    }
    protected val rankCountList = handCalcList map(x => x._1)
    val handVal : HandVal = {
      if (rankCountList == List(1, 1, 1, 1, 1) && ranksAreConsecutive() && isFlush() )
    	HandVal.STRAIGHTFLUSH
      else if (rankCountList == List(4, 1))
        HandVal.POKER
      else if (rankCountList == List(3, 2))
        HandVal.FULLHOUSE
      else if (isFlush())
        HandVal.FLUSH
      else if (rankCountList == List(1, 1, 1, 1, 1) && ranksAreConsecutive() )
    	HandVal.STRAIGHT
      else if (rankCountList == List(3, 1, 1))
        HandVal.THREEOFAKIND
      else if (rankCountList == List(2, 2, 1))
        HandVal.TWOPAIRS
      else if (rankCountList == List(2, 1, 1, 1))
        HandVal.ONEPAIR
      else if (rankCountList == List(1, 1, 1, 1, 1))
        HandVal.HIGHCARD
      else 
        HandVal.NOTVALID
    }
    protected def ranksAreConsecutive() : Boolean = {
      (handCalcList.head._2.id - handCalcList(4)._2.id) == 4 ||
      ( (handCalcList(1)._2.id - handCalcList(4)._2.id) == 3  && handCalcList(0)._2 == Rank.ACE )	// A, 5, 4, 3, 2
    }
    protected def isFlush() : Boolean = {
       val suitList = cards map(x => x._2)
       val l= suitList map(x=> suitList.count(y=> y == x))  // list with no. of occurences of particular suit; in case of Flush: (5, 5, 5, 5, 5)
       l.first == 5
    }
    def compare(that : Hand) : Int = {
       if (this.handVal.id == that.handVal.id) {
         for (i <- 0 until rankCountList.length) {
           if (this.rankCountList(i) < that.rankCountList(i))
             return -1
           if (this.rankCountList(i) > that.rankCountList(i))
             return 1           
         }
         return 0
       } else
         this.handVal.id - that.handVal.id
    }
    override def toString() = {
		cardsAsStr + "     >> " + handVal
    }
}